package convexhull;
/*
 * File: ConvexHull.java
 * ------------------
 * This program is a stub for designing and testing
 * a convex hull algorithm.
 */

import acm.program.*;
import acm.util.*;

import java.awt.*;

@SuppressWarnings("serial")
public class ConvexHull extends ConsoleProgram {
	
	private static final int WIDTH = 300;
	private static final int HEIGHT = 400;

	public void run() {
		
		// Make a display similar to Hangman:
		// console in the left half, graphics in the right
		resize(2 * WIDTH,HEIGHT);
		canvas = new ConvexHullGraphics();
		setLayout(new GridLayout(1, 2));
		add(canvas);
		validate();

		println("This program tests a convex hull method.");
		println();

		while (true) {

			numPoints = readInt("Enter the number of points: ");

			int numRuns = readInt("Number of times to run the algorithm on that many elements: ");
			
			debug = (readLine("Debug? ").toUpperCase().charAt(0) == 'Y');


			totalCHTime = 0;

			for (int i = 0; i < numRuns; i++) {
				// Generate the right number of random points
				initializeDataArray();
				
				// Show the points only in debug mode
				if (debug) {
					canvas.reset();
					canvas.showPoints(dataArray, numPoints);
				}
				
				// Find the convex hull, timing the algorithm
				convexHull();
				
				// Show the result and wait for a prompt only in debug mode
				if (debug) {
					canvas.showHull(dataArray, hullPoints, hullIndices);
					yesOrNo = readLine("Type a carriage-return to continue.");
				}
			}

			println();
			println("Done.");
			println("The algorithm took " + totalCHTime + " milliseconds.");

			yesOrNo = readLine("Again? ").toUpperCase();
			if (yesOrNo.charAt(0) == 'N') break;
		}
		println("Ending execution.");
	}
	
	// Initialize dataArray[0..numPoints-1][0..1] with random integers
	// to the canvas dimensions -- the points are placed so that they
	// lie entirely in the canvas outline
	private void initializeDataArray() {
		dataArray = new int[numPoints][2];
		for (int i = 0; i < numPoints; i++) {
			dataArray[i][0] = rgen.nextInt(canvas.margin() + canvas.pointRadius(), 
					                       canvas.width() - 1 - canvas.margin() - canvas.pointRadius());
			dataArray[i][1] = rgen.nextInt(canvas.margin() + canvas.pointRadius(), 
					                       canvas.height() - 1 - canvas.margin() - canvas.pointRadius());			
		}
	}

	// Compute the convex hull, timing the process
	// Input: int numPoints, int[numPoints][2] dataArray int[numPoints] dataIndices
	// Output: int hullPoints, int[hullPoints] hullIndices
	// This stub routine merely copies all the data over
	// and then sorts them by x-coordinate, for no particular reason
	private void convexHull() {
		double startTime = System.currentTimeMillis();

		// This code should be replaced by a real convex hull algorithm
		hullPoints = numPoints;
		hullIndices = new int[hullPoints];
		for (int i = 0; i < hullPoints; i++) {
			hullIndices[i] = i;
		}

		quickSort(hullIndices, 0, hullPoints - 1);
		
		double ellapsedTime = System.currentTimeMillis() - startTime;
		totalCHTime += ellapsedTime;

	}
	
	// QuickSort routine, using the "key" and "swap" routines to follow
	// This will need modification for a convex hull algorithm
	private void quickSort(int[] indices, int minIndex, int maxIndex) {
		
		int pivot = key(indices, minIndex);
		int i = minIndex + 1;
		int k = maxIndex;
		while (i < k) {
			while ((i <= maxIndex) && (key(indices, i) <= pivot)) i++;
			while ((k >= minIndex) && (key(indices, k) >= pivot)) k--;
			if ((i <= maxIndex) && (k >= minIndex) && (k > i)) swap(indices, i, k);
		}
		if (k >= minIndex) {
			if (key(indices, k) < pivot) swap(indices, minIndex, k);
		} else {
			k = minIndex;
		}
		if (minIndex < k - 1) quickSort(indices, minIndex, k - 1);
		if (k + 1 < maxIndex) quickSort(indices, k + 1, maxIndex);
	}
	
	// For now this routine returns the x-coordinate of the ith point
	private int key(int[] indices, int i) {
		return dataArray[indices[i]][0];
	}
	
	// Swaps the ith and jth points in the array data
	private void swap(int[] indices, int i, int j) {
		int temp;
		temp = indices[i]; indices[i] = indices[j]; indices[j] = temp;
	}


	
	// Instance variables
	
	private int numPoints;
	private int[][] dataArray;
	
	private int numRuns;
	
	private boolean debug;
	
	private int hullPoints;
	private int[] hullIndices;
	
	private double totalCHTime;
	
	private String yesOrNo;

	private RandomGenerator rgen = new RandomGenerator();
	
	private ConvexHullGraphics canvas;
}
