package quicksort;

import acm.program.*;
import acm.util.RandomGenerator;

@SuppressWarnings("serial")
public class BinarySearch extends ConsoleProgram {

	public void run() {

		println("This program times searching methods.");
		println();

		while (true) {

			sortLength = readInt("Enter the number of elements to search: ");
			dataArray = new int[sortLength];
			maxDataEntry = 3 * sortLength;

			int numRuns = readInt("Enter the number times to search that many elements: ");
			
			debug = readLine("Linear search debug/binary search debug? ").toUpperCase();
			lsDebug = (debug.charAt(0) == 'Y');
			bsDebug = (debug.charAt(1) == 'Y');


			totalLSTime = 0;
			totalBSTime = 0;

			for (int i = 0; i < numRuns; i++) {
				initializeDataArray();
				
				if (lsDebug || bsDebug) {
					key = readInt("Value to search for: ");
				} else {
					key = rgen.nextInt(0, maxDataEntry);
				}
				linearSearch();
				binarySearch();
			}

			println();
			println("Done searching.");
			println("The linear searching took " + totalLSTime + " milliseconds.");
			println("The binary searching took " + totalBSTime + " milliseconds.");

			yesOrNo = readLine("Again? ").toUpperCase();
			if (yesOrNo.charAt(0) == 'N')
				break;
		}
		println("Ending execution.");
	}
	
	private void linearSearch() {
		double startTime = System.currentTimeMillis();
	
		linear(0, sortLength - 1);
		
		double ellapsedTime = System.currentTimeMillis() - startTime;
		totalLSTime += ellapsedTime;

		if ((lsDebug) && (keyIndex != -1)) 
			println("Linear search found " + key + " at entry " + keyIndex + ".");
	}
	
	private void linear(int minIndex, int maxIndex) {
		keyIndex = -1;
		for (int i = minIndex; i <= maxIndex; i++)
			if (dataArray[i] == key) keyIndex = i;
	}
	
	private void binarySearch() {
		double startTime = System.currentTimeMillis();
		
		keyIndex = -1;
		binary(0, sortLength - 1);

		double ellapsedTime = System.currentTimeMillis() - startTime;
		totalBSTime += ellapsedTime;

		if ((bsDebug) && (keyIndex != -1)) 
			println("Binary search found " + key + " at entry " + keyIndex + ".");
	}
	
	private void binary(int minIndex, int maxIndex) {
		int midIndex = (minIndex + maxIndex) / 2;
		if (dataArray[midIndex] == key) {
			keyIndex = midIndex;
		} else if ((minIndex < midIndex) && (midIndex < maxIndex) && (dataArray[midIndex] > key)) {
				binary(minIndex, midIndex);
		} else if ((minIndex < midIndex) && (midIndex < maxIndex) && (dataArray[midIndex] < key)) {
				binary(midIndex, maxIndex);
		}
	}

	// Check whether an array is sorted
	private void checkIfSorted(int[] array) {
		for (int i = 1; i < sortLength; i++)
			if (array[i] < array[i - 1]) println("Error in sort.");
	}
	
	// Display the data in an array
	private void showData(int[] array) {
		for (int i = 0; i < sortLength; i++)
			print(" " + array[i]);
		println();
	}
	
	// Initialize dataArray[0..sortLength-1] with random integers
	private void initializeDataArray() {
		for (int i = 0; i < sortLength; i++)
			dataArray[i] = rgen.nextInt(0, maxDataEntry);
		quick(0, sortLength - 1);
		if (lsDebug || bsDebug) {
			checkIfSorted(dataArray);
			showData(dataArray);			
		}
	}

	private void quick(int minIndex, int maxIndex) {
		int temp;
		int pivot = dataArray[minIndex];
		int i = minIndex + 1;
		int k = maxIndex;
		while (i < k) {
			while ((i <= maxIndex) && (dataArray[i] <= pivot)) i++;
			while ((k >= minIndex) && (dataArray[k] >= pivot)) k--;
			if ((i <= maxIndex) && (k >= minIndex) && (k > i)) {
				temp = dataArray[i];
				dataArray[i] = dataArray[k];
				dataArray[k] = temp;
			}
		}
		if (k >= minIndex) {
			if (dataArray[k] < pivot) {
				dataArray[minIndex] = dataArray[k];
				dataArray[k] = pivot;
			}
		} else {
			k = minIndex;
		}
		if (minIndex < k - 1) quick(minIndex, k - 1);
		if (k + 1 < maxIndex) quick(k + 1, maxIndex);
	}


	
	private int[] dataArray;

	private int sortLength;

	private int maxDataEntry;
	
	private int key, keyIndex;
	
	long totalLSTime, totalBSTime;

	private String debug, yesOrNo;
	
	private boolean lsDebug, bsDebug;

	private RandomGenerator rgen = new RandomGenerator();

}
