package quicksort;

import acm.program.*;
import acm.util.RandomGenerator;

@SuppressWarnings("serial")
public class QuickSort extends ConsoleProgram {

	public void run() {

		println("This program times sorting methods.");
		println();

		while (true) {

			sortLength = readInt("Enter the number of elements to sort (more than 1): ");
			dataArray = new int[sortLength];
			maxDataEntry = 3 * sortLength;

			int numRuns = readInt("Enter the number times to sort that many elements: ");
			
			debug = readLine("Bubble sort debug/QuickSort debug? ").toUpperCase();
			bubbleDebug = (debug.charAt(0) == 'Y');
			qsDebug = (debug.charAt(1) == 'Y');


			totalBubbleTime = 0;
			totalQuickTime = 0;

			for (int i = 0; i < numRuns; i++) {
				initializeDataArray();
				bubbleSort();
				quickSort();
			}

			println();
			println("Done sorting.");
			println("The bubble sorting took " + totalBubbleTime + " milliseconds.");
			println("The QuickSorting took " + totalQuickTime + " milliseconds.");

			yesOrNo = readLine("Again? ").toUpperCase();
			if (yesOrNo.charAt(0) == 'N')
				break;
		}
		println("Ending execution.");
	}

	// Initialize dataArray[0..sortLength-1] with random integers
	private void initializeDataArray() {
		for (int i = 0; i < sortLength; i++)
			dataArray[i] = rgen.nextInt(0, maxDataEntry);
	}

	// Display the data in an array
	private void showData(int[] array) {
		for (int i = 0; i < sortLength; i++)
			print(" " + array[i]);
		println();
	}
	
	// Check whether an array is sorted
	private void checkIfSorted(int[] array) {
		for (int i = 1; i < sortLength; i++)
			if (array[i] < array[i - 1]) println("Error in sort.");
	}
	

	private void bubbleSort() {
		int[] bubbleSortArray = new int[sortLength];
		for (int i = 0; i < sortLength; i++) bubbleSortArray[i] = dataArray[i];

		if (bubbleDebug) {println(); print("Before bubble sort: "); showData(bubbleSortArray);}

		double startTime = System.currentTimeMillis();
		int temp;
		for (int i = sortLength - 1; i > 0; i--)
			for (int j = 1; j <= i; j++)
				if (bubbleSortArray[j] < bubbleSortArray[j - 1]) {
					temp = bubbleSortArray[j];
					bubbleSortArray[j] = bubbleSortArray[j - 1];
					bubbleSortArray[j - 1] = temp;
				}
		double ellapsedTime = System.currentTimeMillis() - startTime;
		totalBubbleTime += ellapsedTime;
		
		if (bubbleDebug) {
			print(" After bubble sort: "); 
			showData(bubbleSortArray);
			checkIfSorted(bubbleSortArray);
		}

	}

	private void quickSort() {
		quickSortArray = new int[sortLength];
		for (int i = 0; i < sortLength; i++) quickSortArray[i] = dataArray[i];

		if (qsDebug) {println(); print("Before QuickSort: "); showData(quickSortArray);}

		double startTime = System.currentTimeMillis();

		quick(0,sortLength - 1);

		double ellapsedTime = System.currentTimeMillis() - startTime;
		totalQuickTime += ellapsedTime;
		
		if (qsDebug) {
			print(" After QuickSort: "); 
			showData(quickSortArray);
			checkIfSorted(quickSortArray);
		}
	}
	
	// This routine assumes that 0 <= minIndex < maxIndex <= sortLength-1
	private void quick(int minIndex, int maxIndex) {
		if (qsDebug) {
			print(" Quick called to sort: "); showData(quickSortArray);
			println("   with parameters " + minIndex + ", " + maxIndex + ".");
		}
		int temp;
		int pivot = quickSortArray[minIndex];
		int i = minIndex + 1;
		int k = maxIndex;
		while (i < k) {
			while ((i <= maxIndex) && (quickSortArray[i] <= pivot)) i++;
			while ((k >= minIndex) && (quickSortArray[k] >= pivot)) k--;
			if ((i <= maxIndex) && (k >= minIndex) && (k > i)) {
//				println(" Exchanging entries " + i + " and " + k + ".");
				temp = quickSortArray[i];
				quickSortArray[i] = quickSortArray[k];
				quickSortArray[k] = temp;
			}
		}
		if (k >= minIndex) {
			if (quickSortArray[k] < pivot) {
				quickSortArray[minIndex] = quickSortArray[k];
				quickSortArray[k] = pivot;
			}
		} else {
			k = minIndex;
		}
//		println(" k=" + k + ".");
		if (qsDebug) {
			print(" The sort produced: "); showData(quickSortArray);
		}
//		println("Next I would consider sorting from " + minIndex + " to " + (k-1)
//				+ " and from " + (k+1) + " to " + maxIndex + ".");
		if (minIndex < k - 1) quick(minIndex, k - 1);
		if (k + 1 < maxIndex) quick(k + 1, maxIndex);
	}


	
	private int[] dataArray, quickSortArray;

	private int sortLength;

	private int maxDataEntry;
	
	long totalBubbleTime, totalQuickTime;

	private String debug, yesOrNo;
	
	private boolean bubbleDebug, qsDebug;

	private RandomGenerator rgen = new RandomGenerator();

}
